# Future
from __future__ import annotations


# isort: off
from spotipy.objects.base import *
# isort: on

# My stuff
from spotipy.objects.actions import *
from spotipy.objects.album import *
from spotipy.objects.artist import *
from spotipy.objects.category import *
from spotipy.objects.copyright import *
from spotipy.objects.device import *
from spotipy.objects.enums import *
from spotipy.objects.episode import *
from spotipy.objects.followers import *
from spotipy.objects.image import *
from spotipy.objects.playback import *
from spotipy.objects.playlist import *
from spotipy.objects.recommendation import *
from spotipy.objects.search import *
from spotipy.objects.show import *
from spotipy.objects.token import *
from spotipy.objects.track import *
from spotipy.objects.user import *



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import CategoryData


__all__ = (
    "Category",
)


class Category:

    def __init__(self, data: CategoryData) -> None:
        self.href = data["href"]
        self.icons = [objects.Image(image) for image in data["icons"]]
        self.id = data["id"]
        self.name = data["name"]

    def __repr__(self) -> str:
        return f"<spotipy.Category id='{self.id}' name='{self.name}'>"



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import ArtistData, SimpleArtistData


__all__ = (
    "SimpleArtist",
    "Artist",
)


class SimpleArtist(objects.BaseObject):

    def __init__(self, data: SimpleArtistData) -> None:
        super().__init__(data)

        self.external_urls = data["external_urls"]

    def __repr__(self) -> str:
        return f"<spotipy.SimpleArtist id='{self.id}', name='{self.name}'>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")


class Artist(SimpleArtist):

    def __init__(self, data: ArtistData) -> None:
        super().__init__(data)

        self.external_urls = data["external_urls"]
        self.followers = objects.Followers(data["followers"])
        self.genres = data["genres"]
        self.images = [objects.Image(image) for image in data["images"]]
        self.popularity = data["popularity"]

    def __repr__(self) -> str:
        return f"<spotipy.Artist id='{self.id}', name='{self.name}'>"



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import CopyrightData


__all__ = (
    "Copyright",
)


class Copyright:

    def __init__(self, data: CopyrightData) -> None:
        self.text = data["text"]
        self.type = objects.CopyrightType(data["type"])

    def __repr__(self) -> str:
        return f"<spotipy.Copyright type={self.type!r}>"



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import ContextData, CurrentlyPlayingData, PlaybackStateData


__all__ = (
    "Context",
    "PlaybackState",
    "CurrentlyPlaying",
)


class Context:

    def __init__(self, data: ContextData) -> None:
        self.external_urls = data["external_urls"]
        self.href = data["href"]
        self.type = data["type"]
        self.uri = data["uri"]

    def __repr__(self) -> str:
        return "<spotipy.Context"


class PlaybackState:

    def __init__(self, data: PlaybackStateData) -> None:
        self.actions = objects.Actions(data["actions"])
        self.context = Context(data["context"])
        self.currently_playing_type = data["currently_playing_type"]
        self.device = objects.Device(data["device"])
        self.is_playing = data["is_playing"]
        self.item = objects.Track(item) if (item := data["item"]) else None
        self.progress_ms = data["progress_ms"]
        self.repeat_state = data["repeat_state"]
        self.shuffle_state = data["shuffle_state"]
        self.timestamp: int = data["timestamp"]

    def __repr__(self) -> str:
        return "<spotipy.CurrentlyPlayingContext>"


class CurrentlyPlaying:

    def __init__(self, data: CurrentlyPlayingData) -> None:
        self.context = Context(data["context"])
        self.currently_playing_type = data["currently_playing_type"]
        self.is_playing = data["is_playing"]
        self.item = objects.Track(item) if (item := data["item"]) else None
        self.progress_ms = data["progress_ms"]
        self.timestamp: int = data["timestamp"]

    def __repr__(self) -> str:
        return "<spotipy.CurrentlyPlaying>"



# Future
from __future__ import annotations

# Standard Library
import logging
from typing import Final, Literal, NamedTuple

# My stuff
from .client import *
from .exceptions import *
from .http import *
from .objects import *
from .utils import *
from .values import *


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: Literal["alpha", "beta", "candidate", "final"]
    serial: int


version_info: Final[VersionInfo] = VersionInfo(major=0, minor=1, micro=1, releaselevel="final", serial=0)

__title__: Final[str] = "spotipy"
__author__: Final[str] = "Axelancerr"
__copyright__: Final[str] = "Copyright 2021-present Axelancerr"
__license__: Final[str] = "MIT"
__version__: Final[str] = "0.1.1"
__maintainer__: Final[str] = "Aaron Hennessey"
__source__: Final[str] = "https://github.com/Axelware/spoti.py"

logging.getLogger("spotipy")



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import RecommendationData, RecommendationSeedData


__all__ = (
    "RecommendationSeed",
    "Recommendation"
)


class RecommendationSeed:

    def __init__(self, data: RecommendationSeedData) -> None:
        self.initial_pool_size = data["initialPoolSize"]
        self.after_filtering_size = data["afterFilteringSize"]
        self.after_relinking_size = data["afterRelinkingSize"]
        self.href = data["href"]
        self.id = data["id"]
        self.type = data["type"]

    def __repr__(self) -> str:
        return f"<spotipy.RecommendationSeed id='{self.id}', type='{self.type}'>"


class Recommendation:

    def __init__(self, data: RecommendationData) -> None:
        self.tracks = [objects.Track(data) for data in data["tracks"]]
        self.seeds = [objects.RecommendationSeed(data) for data in data["seeds"]]

    def __repr__(self) -> str:
        return f"<spotipy.Recommendation tracks={len(self.tracks)}, seeds={len(self.seeds)}>"



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import SearchResultData


__all__ = (
    "SearchResult",
)


class SearchResult:

    def __init__(self, data: SearchResultData) -> None:
        self._albums_paging: objects.PagingObject | None = objects.PagingObject(paging) if (paging := data["albums"]) else None
        self._artists_paging: objects.PagingObject | None = objects.PagingObject(paging) if (paging := data["artists"]) else None
        self._playlists_paging: objects.PagingObject | None = objects.PagingObject(paging) if (paging := data["playlists"]) else None
        self._tracks_paging: objects.PagingObject | None = objects.PagingObject(paging) if (paging := data["tracks"]) else None
        self._shows_paging: objects.PagingObject | None = objects.PagingObject(paging) if (paging := data["shows"]) else None
        self._episodes_paging: objects.PagingObject | None = objects.PagingObject(paging) if (paging := data["episodes"]) else None

        self.albums: list[objects.SimpleAlbum | None] = [
            objects.SimpleAlbum(album) for album in self._albums_paging.items
        ] if self._albums_paging else []

        self.artists: list[objects.Artist | None] = [
            objects.Artist(artist) for artist in self._artists_paging.items
        ] if self._artists_paging else []

        self.playlists: list[objects.SimplePlaylist | None] = [
            objects.SimplePlaylist(playlist) for playlist in self._playlists_paging.items
        ] if self._playlists_paging else []

        self.tracks: list[objects.Track | None] = [
            objects.Track(track) for track in self._tracks_paging.items
        ] if self._tracks_paging else []

        #  self.shows = ...

        #  self.episodes = ...

    def __repr__(self) -> str:
        return f"<spotipy.SearchResult albums={len(self.albums)}, artists={len(self.artists)}, playlists={len(self.playlists)}, tracks={len(self.tracks)}>"



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import AlbumData, AlbumRestrictionData, SimpleAlbumData


__all__ = (
    "AlbumRestriction",
    "SimpleAlbum",
    "Album"
)


class AlbumRestriction:

    def __init__(self, data: AlbumRestrictionData) -> None:
        self.reason = data["reason"]

    def __repr__(self) -> str:
        return f"<spotipy.AlbumRestriction reason='{self.reason}'>"


class SimpleAlbum(objects.BaseObject):

    def __init__(self, data: SimpleAlbumData) -> None:
        super().__init__(data)

        self.album_type = data["album_type"]
        self.artists = [objects.SimpleArtist(artist) for artist in data["artists"]]
        self.available_markets = data.get("available_markets")
        self.external_urls = data["external_urls"]
        self.images = [objects.Image(image) for image in data["images"]]
        self.release_date = data["release_date"]
        self.release_data_precision = data["release_date_precision"]
        self.restriction = AlbumRestriction(restriction) if (restriction := data.get("restrictions")) else None
        self.total_tracks = data.get("total_tracks")

    def __repr__(self) -> str:
        return f"<spotipy.SimpleAlbum id='{self.id}', name='{self.name}', total_tracks={self.total_tracks}>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")


class Album(objects.BaseObject):

    def __init__(self, data: AlbumData) -> None:
        super().__init__(data)

        self.album_type = data["album_type"]
        self.artists = [objects.SimpleArtist(artist) for artist in data["artists"]]
        self.available_markets = data.get("available_markets")
        self.copyrights = [objects.Copyright(copyright) for copyright in data["copyrights"]]
        self.external_ids = data["external_ids"]
        self.external_urls = data["external_urls"]
        self.genres = data["genres"]
        self.images = [objects.Image(image) for image in data["images"]]
        self.label = data["label"]
        self.popularity = data["popularity"]
        self.release_date = data["release_date"]
        self.release_data_precision = data["release_date_precision"]
        self.restriction = AlbumRestriction(restriction) if (restriction := data.get("restrictions")) else None
        self.total_tracks = data["total_tracks"]

        self._tracks_paging = objects.PagingObject(data["tracks"])
        self.tracks = [objects.SimpleTrack(track) for track in self._tracks_paging.items]

    def __repr__(self) -> str:
        return f"<spotipy.Album id='{self.id}', name='{self.name}', total_tracks={self.total_tracks}>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")



# Future
from __future__ import annotations

# Standard Library
from typing import Any

# Packages
import aiohttp


__all__ = (
    "SpotifyException",
    "AuthenticationError",
    "HTTPError",
    "BadRequest",
    "Unauthorized",
    "Forbidden",
    "NotFound",
    "RequestEntityTooLarge",
    "TooManyRequests",
)


class SpotifyException(Exception):
    pass


class AuthenticationError(SpotifyException):

    def __init__(
        self,
        response: aiohttp.ClientResponse,
        /,
        *,
        data: dict[str, Any]
    ) -> None:

        self._response: aiohttp.ClientResponse = response
        self._error: str = data["error"]
        self._error_description: str = data["error_description"]

    @property
    def response(self) -> aiohttp.ClientResponse:
        return self._response

    @property
    def error(self) -> str:
        return self._error

    @property
    def error_description(self) -> str:
        return self._error_description


class HTTPError(SpotifyException):

    def __init__(
        self,
        response: aiohttp.ClientResponse,
        /,
        *,
        data: dict[str, dict[str, Any]] | None
    ) -> None:

        self._response: aiohttp.ClientResponse = response
        self._status: int = response.status

        if isinstance(data, dict):
            self._message: str = data["error"]["message"]
        else:
            self._message: str = data or ""

        message = f"{self.response.status} {self.response.reason}"
        if len(self.message):
            message += f": {self.message}"

        super().__init__(message)

    @property
    def response(self) -> aiohttp.ClientResponse:
        return self._response

    @property
    def status(self) -> int:
        return self._status

    @property
    def message(self) -> str:
        return self._message


class BadRequest(HTTPError):
    pass


class Unauthorized(HTTPError):
    pass


class Forbidden(HTTPError):
    pass


class NotFound(HTTPError):
    pass


class RequestEntityTooLarge(HTTPError):
    pass


class TooManyRequests(HTTPError):
    pass


class SpotifyServerError(HTTPError):
    pass



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import EpisodeData, EpisodeRestrictionData, EpisodeResumePointData, SimpleEpisodeData


__all__ = (
    "EpisodeRestriction",
    "EpisodeResumePoint",
    "SimpleEpisode",
    "Episode",
)


class EpisodeRestriction:

    def __init__(self, data: EpisodeRestrictionData) -> None:
        self.reason = data["reason"]

    def __repr__(self) -> str:
        return f"<spotipy.EpisodeRestriction reason='{self.reason}'>"


class EpisodeResumePoint:

    def __init__(self, data: EpisodeResumePointData) -> None:
        self.fully_played = data["fully_played"]
        self.resume_position_ms = data["resume_position_ms"]

    def __repr__(self) -> str:
        return f"<spotipy.EpisodeResumePoint fully_played={self.fully_played} resume_position_ms={self.resume_position_ms}>"


class SimpleEpisode(objects.BaseObject):

    def __init__(self, data: SimpleEpisodeData) -> None:
        super().__init__(data)

        self.audio_preview_url = data["audio_preview_url"]
        self.description = data["description"]
        self.duration_ms = data["duration_ms"]
        self.explicit = data["explicit"]
        self.external_urls = data["external_urls"]
        self.html_description = data["html_description"]
        self.images = [objects.Image(image) for image in data["images"]]
        self.is_externally_hosted = data["is_externally_hosted"]
        self.is_playable = data["is_playable"]
        self.languages = data["languages"]
        self.release_date = data["release_date"]
        self.release_date_precision = data["release_date_precision"]
        self.restriction = EpisodeRestriction(restriction) if (restriction := data.get("restrictions")) else None
        self.resume_point = EpisodeResumePoint(resume_point) if (resume_point := data.get("resume_point")) else None

    def __repr__(self) -> str:
        return f"<spotipy.SimpleEpisode id='{self.id}', name='{self.name}'>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")


class Episode(objects.BaseObject):

    def __init__(self, data: EpisodeData) -> None:
        super().__init__(data)

        self.audio_preview_url = data["audio_preview_url"]
        self.description = data["description"]
        self.duration_ms = data["duration_ms"]
        self.explicit = data["explicit"]
        self.external_urls = data["external_urls"]
        self.html_description = data["html_description"]
        self.images = [objects.Image(image) for image in data["images"]]
        self.is_externally_hosted = data["is_externally_hosted"]
        self.is_playable = data["is_playable"]
        self.languages = data["languages"]
        self.release_date = data["release_date"]
        self.release_date_precision = data["release_date_precision"]
        self.show = objects.Show(data["show"])
        self.restriction = EpisodeRestriction(restriction) if (restriction := data.get("restrictions")) else None
        self.resume_point = EpisodeResumePoint(resume_point) if (resume_point := data.get("resume_point")) else None

    def __repr__(self) -> str:
        return f"<spotipy.Episode id='{self.id}', name='{self.name}'>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")



# Future
from __future__ import annotations

# My stuff
from spotipy.typings.objects import DeviceData


__all__ = (
    "Device",
)


class Device:

    def __init__(self, data: DeviceData) -> None:
        self.id = data["id"]
        self.is_active = data["is_active"]
        self.is_private_session = data["is_private_session"]
        self.is_restricted = data["is_restricted"]
        self.name = data["name"]
        self.type = data["type"]
        self.volume_percent = data["volume_percent"]

    def __repr__(self) -> str:
        return f"<spotipy.Device id='{self.id}' name='{self.name}'>"



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import PlaylistData, SimplePlaylistData


__all__ = (
    "SimplePlaylist",
    "Playlist"
)


class SimplePlaylist(objects.BaseObject):

    def __init__(self, data: SimplePlaylistData) -> None:
        super().__init__(data)

        self.collaborative = data["collaborative"]
        self.description = data["description"]
        self.external_urls = data["external_urls"]
        self.images = [objects.Image(image) for image in data["images"]]
        self.owner = objects.User(data["owner"])
        self.primary_color = data["primary_color"]
        self.public = data["public"]
        self.snapshot_id = data["snapshot_id"]
        self.tracks_href = data["tracks"]["href"]
        self.total_tracks = data["tracks"]["total"]

    def __repr__(self) -> str:
        return f"<spotipy.SimplePlaylist id='{self.id}', name='{self.name}', total_tracks={self.total_tracks}>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")


class Playlist(objects.BaseObject):

    def __init__(self, data: PlaylistData) -> None:
        super().__init__(data)

        self.collaborative = data["collaborative"]
        self.description = data["description"]
        self.external_urls = data["external_urls"]
        self.followers = objects.Followers(data["followers"])
        self.images = [objects.Image(image) for image in data["images"]]
        self.owner = objects.User(data["owner"])
        self.primary_color = data["primary_color"]
        self.public = data["public"]
        self.snapshot_id = data["snapshot_id"]
        self.total_tracks = data["tracks"]["total"]

        self._tracks_paging = objects.PagingObject(data["tracks"])
        self.tracks = [objects.PlaylistTrack(track) for track in self._tracks_paging.items]

    def __repr__(self) -> str:
        return f"<spotipy.Playlist id='{self.id}', name='{self.name}', total_tracks={self.total_tracks}>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")



# Future
from __future__ import annotations

# My stuff
from spotipy.typings.objects import FollowersData


__all__ = (
    "Followers",
)


class Followers:

    def __init__(self, data: FollowersData) -> None:
        self.href = data["href"]
        self.total = data["total"]

    def __repr__(self) -> str:
        return f"<spotipy.Followers total={self.total}>"



# Future
from __future__ import annotations

# Standard Library
import time

# Packages
import aiohttp

# My stuff
from spotipy import exceptions, values
from spotipy.typings.objects import ClientCredentialsData, UserCredentialsData


__all__ = (
    "ClientCredentials",
    "UserCredentials"
)


class ClientCredentials:

    def __init__(self, data: ClientCredentialsData, client_id: str, client_secret: str) -> None:

        self._access_token = data["access_token"]
        self._token_type = data["token_type"]
        self._expires_in = data["expires_in"]

        self._client_id: str = client_id
        self._client_secret: str = client_secret

        self._last_authorized_time: float = time.time()

    def __repr__(self) -> str:
        return "<spotipy.ClientCredentials>"

    #

    @property
    def access_token(self) -> str:
        return self._access_token

    @property
    def token_type(self) -> str:
        return self._token_type

    @property
    def expires_in(self) -> int:
        return self._expires_in

    #

    def is_expired(self) -> bool:
        return (time.time() - self._last_authorized_time) >= self.expires_in

    async def refresh(
        self,
        *,
        session: aiohttp.ClientSession
    ) -> None:

        data = {
            "grant_type":    "client_credentials",
            "client_id":     self._client_id,
            "client_secret": self._client_secret
        }

        async with session.post(url=values.TOKEN_URL, data=data) as response:

            data = await response.json()

            if data.get("error"):
                raise exceptions.AuthenticationError(response, data=data)

            self._access_token = data["access_token"]
            self._token_type = data["token_type"]
            self._expires_in = data["expires_in"]

            self._last_authorized_time = time.time()

    @classmethod
    async def from_client_secret(
        cls,
        client_id: str,
        client_secret: str,
        *,
        session: aiohttp.ClientSession
    ) -> ClientCredentials:

        data = {
            "grant_type":    "client_credentials",
            "client_id":     client_id,
            "client_secret": client_secret
        }

        async with session.post(url=values.TOKEN_URL, data=data) as response:

            data = await response.json()

            if data.get("error"):
                raise exceptions.AuthenticationError(response, data=data)

            return cls(data, client_id=client_id, client_secret=client_secret)


class UserCredentials:

    def __init__(self, data: UserCredentialsData, client_id: str, client_secret: str) -> None:

        self._access_token = data["access_token"]
        self._token_type = data["token_type"]
        self._scope = data["scope"]
        self._expires_in = data["expires_in"]
        self._refresh_token = data.get("refresh_token")

        self._client_id: str = client_id
        self._client_secret: str = client_secret

        self._last_authorized_time: float = time.time()

    def __repr__(self) -> str:
        return "<spotipy.UserCredentials>"

    #

    @property
    def access_token(self) -> str:
        return self._access_token

    @property
    def token_type(self) -> str:
        return self._token_type

    @property
    def scope(self) -> str:
        return self._scope

    @property
    def expires_in(self) -> int:
        return self._expires_in

    #

    def is_expired(self) -> bool:
        return (time.time() - self._last_authorized_time) >= self.expires_in

    async def refresh(
        self,
        *,
        session: aiohttp.ClientSession
    ) -> None:

        data = {
            "grant_type":    "refresh_token",
            "refresh_token": self._refresh_token,
            "client_id":     self._client_id,
            "client_secret": self._client_secret
        }

        async with session.post(url=values.TOKEN_URL, data=data) as response:

            data = await response.json()

            if data.get("error"):
                raise exceptions.AuthenticationError(response, data=data)

            self._access_token = data["access_token"]
            self._token_type = data["token_type"]
            self._scope = data["scope"]
            self._expires_in = data["expires_in"]

            self._last_authorized_time = time.time()

    @classmethod
    async def from_refresh_token(
        cls,
        client_id: str,
        client_secret: str,
        *,
        session: aiohttp.ClientSession,
        refresh_token: str,
    ) -> UserCredentials:

        data = {
            "client_id":     client_id,
            "client_secret": client_secret,
            "grant_type":    "refresh_token",
            "refresh_token": refresh_token,
        }

        async with session.post(url=values.TOKEN_URL, data=data) as response:

            data = await response.json()

            if data.get("error"):
                raise exceptions.AuthenticationError(response, data=data)

            data["refresh_token"] = refresh_token

            return cls(data, client_id=client_id, client_secret=client_secret)



# Future
from __future__ import annotations

# My stuff
from spotipy.objects import album, artist, base, enums, user
from spotipy.typings.objects import (
    AudioFeaturesData,
    PlaylistTrackData,
    SimpleTrackData,
    TrackData,
    TrackRestrictionData,
)


__all__ = (
    "AudioFeatures",
    "TrackRestriction",
    "SimpleTrack",
    "Track",
    "PlaylistTrack",
)


class AudioFeatures:

    def __init__(self, data: AudioFeaturesData) -> None:
        self.acousticness = data["acousticness"]
        self.analysis_url = data["analysis_url"]
        self.danceability = data["danceability"]
        self.duration_ms = data["duration_ms"]
        self.energy = data["energy"]
        self.id = data["id"]
        self.instrumentalness = data["instrumentalness"]
        self.key = enums.Key(data["key"])
        self.liveness = data["liveness"]
        self.loudness = data["loudness"]
        self.mode = enums.Mode(data["mode"])
        self.speechiness = data["speechiness"]
        self.tempo = data["tempo"]
        self.time_signature = data["time_signature"]
        self.track_href = data["track_href"]
        self.type = data["type"]
        self.uri = data["uri"]
        self.valence = data["valence"]

    def __repr__(self) -> str:
        return f"<spotipy.AudioFeatures id='{self.id}'>"


class TrackRestriction:

    def __init__(self, data: TrackRestrictionData) -> None:
        self.reason = data["reason"]

    def __repr__(self) -> str:
        return f"<spotipy.TrackRestriction reason='{self.reason}'>"


class SimpleTrack(base.BaseObject):

    def __init__(self, data: SimpleTrackData) -> None:
        super().__init__(data)

        self.artists = [artist.SimpleArtist(artist_data) for artist_data in data["artists"]]
        self.available_markets = data["available_markets"]
        self.disc_number = data["disc_number"]
        self.duration_ms = data["duration_ms"]
        self.explicit = data["explicit"]
        self.external_urls = data["external_urls"]
        self.is_local = data["is_local"]
        self.preview_url = data["preview_url"]
        self.restriction = TrackRestriction(restriction) if (restriction := data.get("restrictions")) else None
        self.track_number = data["track_number"]

    def __repr__(self) -> str:
        return f"<spotipy.SimpleTrack id='{self.id}', name='{self.name}'>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")


class Track(base.BaseObject):

    def __init__(self, data: TrackData) -> None:
        super().__init__(data)

        self.album = album.SimpleAlbum(data["album"])
        self.artists = [artist.SimpleArtist(artist_data) for artist_data in data["artists"]]
        self.available_markets = data.get("available_markets")
        self.disc_number = data["disc_number"]
        self.duration_ms = data["duration_ms"]
        self.explicit = data["explicit"]
        self.external_ids = data["external_ids"]
        self.external_urls = data["external_urls"]
        self.is_local = data["is_local"]
        self.popularity = data["popularity"]
        self.preview_url = data["preview_url"]
        self.restriction = TrackRestriction(restriction) if (restriction := data.get("restrictions")) else None
        self.track_number = data["track_number"]

    def __repr__(self) -> str:
        return f"<spotipy.Track id='{self.id}', name='{self.name}'>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")


class PlaylistTrack(base.BaseObject):

    def __init__(self, data: PlaylistTrackData) -> None:
        super().__init__(data["track"])

        self.added_at = data["added_at"]
        self.added_by = user.User(data["added_by"])
        self.is_local = data["is_local"]
        self.primary_colour = data["primary_color"]
        self.video_thumbnail = data["video_thumbnail"]["url"]

        track = data["track"]
        self.album = album.SimpleAlbum(track["album"])
        self.artists = [artist.SimpleArtist(artist_data) for artist_data in track["artists"]]
        self.available_markets = track.get("available_markets")
        self.disc_number = track["disc_number"]
        self.duration_ms = track["duration_ms"]
        self.explicit = track["explicit"]
        self.external_ids = track["external_ids"]
        self.external_urls = track["external_urls"]
        self.is_local = track["is_local"]
        self.popularity = track["popularity"]
        self.preview_url = track["preview_url"]
        self.restriction = TrackRestriction(restriction) if (restriction := track.get("restrictions")) else None
        self.track_number = track["track_number"]

    def __repr__(self) -> str:
        return f"<spotipy.PlaylistTrack id='{self.id}', name='{self.name}'>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")



# Future
from __future__ import annotations


__all__ = (
    "SCOPES",
    "VALID_SEED_KWARGS",
)


SCOPES = [
    "ugc-image-upload",

    "playlist-modify-private",
    "playlist-read-private",
    "playlist-modify-public",
    "playlist-read-collaborative",

    "user-read-private",
    "user-read-email",

    "user-read-playback-state",
    "user-modify-playback-state",
    "user-read-currently-playing",

    "user-library-modify",
    "user-library-read",

    "user-read-playback-position",
    "user-read-recently-played",
    "user-top-read",

    "app-remote-control",
    "streaming",

    "user-follow-modify",
    "user-follow-read",
]

VALID_SEED_KWARGS = [
    "min_acousticness",
    "max_acousticness",
    "target_acousticness",
    "min_danceability",
    "max_danceability",
    "target_danceability",
    "min_duration_ms",
    "max_duration_ms",
    "target_duration_ms",
    "min_energy",
    "max_energy",
    "target_energy",
    "min_instrumentalness",
    "max_instrumentalness",
    "target_instrumentalness",
    "min_key",
    "max_key",
    "target_key",
    "min_liveness",
    "max_liveness",
    "target_liveness",
    "min_loudness",
    "max_loudness",
    "target_loudness",
    "min_mode",
    "max_mode",
    "target_mode",
    "min_popularity",
    "max_popularity",
    "target_popularity",
    "min_speechiness",
    "max_speechiness",
    "target_speechiness",
    "min_tempo",
    "max_tempo",
    "target_tempo",
    "min_time_signature",
    "max_time_signature",
    "target_time_signature",
    "min_valence",
    "max_valence",
    "target_valence"
]

BASE_URL = "https://api.spotify.com/v1"
TOKEN_URL = "https://accounts.spotify.com/api/token"



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import ShowData


__all__ = (
    "Show",
)


class Show(objects.BaseObject):

    def __init__(self, data: ShowData) -> None:
        super().__init__(data)

        self.available_markets = data.get("available_markets")
        self.copyrights = [objects.Copyright(copyright) for copyright in data["copyrights"]]
        self.description = data["description"]
        self.explicit = data["explicit"]
        self.external_urls = data["external_urls"]
        self.html_description = data["html_description"]
        self.images = [objects.Image(image) for image in data["images"]]
        self.is_externally_hosted = data["is_externally_hosted"]
        self.languages = data["languages"]
        self.media_type = data["media_type"]
        self.publisher = data["publisher"]
        self.total_episodes = data["total_episodes"]

    def __repr__(self) -> str:
        return f"<spotipy.Show id='{self.id}', name='{self.name}'>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")



# Future
from __future__ import annotations

# My stuff
from spotipy.typings.objects import ImageData


__all__ = (
    "Image",
)


class Image:

    def __init__(self, data: ImageData) -> None:
        self.url = data["url"]
        self.width = data["width"]
        self.height = data["height"]

    def __repr__(self) -> str:
        return f"<spotipy.Image url='<{self.url}>', width={self.width}, height={self.height}>"



# Future
from __future__ import annotations

# My stuff
from spotipy import objects
from spotipy.typings.objects import ExplicitContentSettingsData, UserData


__all__ = (
    "ExplicitContentSettings",
    "User"
)


class ExplicitContentSettings:

    def __init__(self, data: ExplicitContentSettingsData) -> None:
        self.filter_enabled = data["filter_enabled"]
        self.filter_locked = data["filter_locked"]

    def __repr__(self) -> str:
        return f"<spotipy.ExplicitContentSettings filter_enabled={self.filter_enabled}, filter_locked={self.filter_locked}"


class User(objects.BaseObject):

    def __init__(self, data: UserData) -> None:
        super().__init__(data)

        self.country = data.get("country")
        self.display_name = data.get("display_name")
        self.email = data.get("email")
        self.explicit_content_settings = ExplicitContentSettings(explicit_content) if (explicit_content := data.get("explicit_content")) else None
        self.external_urls = data["external_urls"]
        self.followers = objects.Followers(followers) if (followers := data.get("followers")) else None
        self.images = [objects.Image(image) for image in images] if (images := data.get("images")) else None
        self.product = data.get("product")

    def __repr__(self) -> str:
        return f"<spotipy.User id='{self.id}', name='{self.display_name}'>"

    #

    @property
    def url(self) -> str | None:
        return self.external_urls.get("spotify")



# Future
from __future__ import annotations

# My stuff
from spotipy.typings.objects import AlternativePagingObjectData, BaseObjectData, PagingObjectData


__all__ = (
    "BaseObject",
    "PagingObject",
)


class BaseObject:

    def __init__(self, data: BaseObjectData) -> None:
        self.href = data["href"]
        self.id = data["id"]
        self.name = data.get("name")
        self.type = data["type"]
        self.uri = data["uri"]

    def __repr__(self) -> str:
        return f"<spotipy.BaseObject id='{self.id}', name='{self.name}'>"


class PagingObject:

    def __init__(self, data: PagingObjectData) -> None:
        self.href = data["href"]
        self.items = data["items"]
        self.limit = data["limit"]
        self.next = data["next"]
        self.offset = data["offset"]
        self.previous = data["previous"]
        self.total = data["total"]

    def __repr__(self) -> str:
        return f"<spotify.PagingObject total={self.total}, limit={self.limit}, offset={self.offset}>"


class AlternativePagingObject:

    def __init__(self, data: AlternativePagingObjectData) -> None:
        self.href = data["href"]
        self.items = data["items"]
        self.limit = data["limit"]
        self.next = data["next"]
        self.after = data["after"]
        self.before = data.get("before")

    def __repr__(self) -> str:
        return f"<spotify.AlternativePagingObject limit={self.limit}, before={self.before}, after={self.after}>"



# Future
from __future__ import annotations

# Standard Library
import asyncio
import base64
import logging
import urllib.parse
from collections.abc import Sequence
from typing import Any

# Packages
import aiohttp

# My stuff
from spotipy import exceptions, objects, typings, utils, values


__all__ = (
    "Route",
    "HTTPClient"
)

__log__: logging.Logger = logging.getLogger("spotipy.http")


class Route:

    def __init__(
        self,
        method: typings.HTTPMethod,
        path: str,
        /,
        **parameters: Any
    ) -> None:

        url = values.BASE_URL + path
        if parameters:
            url = url.format_map({key: urllib.parse.quote(value) if isinstance(value, str) else value for key, value in parameters.items()})

        self.method: typings.HTTPMethod = method
        self.path: str = path
        self.parameters: dict[str, Any] | None = parameters
        self.url: str = url

    def __repr__(self) -> str:
        return f"<spotipy.Route method='{self.method}', url='{self.url}'>"


class HTTPClient:

    def __init__(
        self,
        *,
        client_id: str,
        client_secret: str,
        session: aiohttp.ClientSession | None
    ) -> None:

        self._client_id: str = client_id
        self._client_secret: str = client_secret
        self._session: aiohttp.ClientSession | None = session

        self._client_credentials: objects.ClientCredentials | None = None

    def __repr__(self) -> str:
        return "<spotipy.HTTPClient>"

    #

    async def _get_session(self) -> aiohttp.ClientSession:

        if not self._session or self._session.closed:
            self._session = aiohttp.ClientSession()

        return self._session

    async def _get_credentials(
        self,
        _credentials: objects.ClientCredentials | objects.UserCredentials | None = None,
        /
    ) -> objects.ClientCredentials | objects.UserCredentials:

        if not self._client_credentials:
            self._client_credentials = await objects.ClientCredentials.from_client_secret(
                self._client_id,
                self._client_secret,
                session=self._session
            )

        credentials = _credentials or self._client_credentials

        if credentials.is_expired():
            await credentials.refresh(session=self._session)

        return credentials

    #

    async def request(
        self,
        route: Route,
        /,
        *,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None,
        parameters: dict[str, Any] | None = None,
        data: dict[str, Any] | str | None = None,
        json: dict[str, Any] | None = None
    ) -> Any:

        _session = await self._get_session()
        _credentials = await self._get_credentials(credentials)

        headers = {
            "Authorization": f"Bearer {_credentials.access_token}"
        }

        if json:
            headers["Content-Type"] = "application/json"
            data = utils.to_json(json)

        response: aiohttp.ClientResponse | None = None
        response_data: dict[str, Any] | str | None = None

        for tries in range(3):

            try:

                async with _session.request(
                        method=route.method,
                        url=route.url,
                        headers=headers,
                        params=parameters,
                        data=data
                ) as response:

                    response_data = await utils.json_or_text(response)

                    match response.status:

                        case 200 | 201 | 202 | 204:
                            return response_data

                        case 400:
                            raise exceptions.BadRequest(response, data=response_data)
                        case 401:
                            raise exceptions.Unauthorized(response, data=response_data)
                        case 403:
                            raise exceptions.Forbidden(response, data=response_data)
                        case 404:
                            raise exceptions.NotFound(response, data=response_data)
                        case 413:
                            raise exceptions.RequestEntityTooLarge(
                                response,
                                data={"error": {"status": 413, "message": "Image was too large."}}
                            )

                        case 429:
                            await asyncio.sleep(float(response.headers["Retry-After"]))
                            continue

                        case 500 | 502 | 503:
                            await asyncio.sleep(1 + tries * 2)
                            continue

            except OSError as error:
                if tries < 2 and error.errno in (54, 10054):
                    await asyncio.sleep(1 + tries * 2)
                    continue
                raise

        if response:

            if response.status >= 500:
                raise exceptions.SpotifyServerError(response, data=response_data)

            raise exceptions.HTTPError(response, data=response_data)

        raise RuntimeError("This shouldn't happen.")

    # ALBUMS API

    async def get_album(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.AlbumData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/albums/{id}", id=_id), parameters=parameters, credentials=credentials)

    async def get_albums(
        self,
        ids: Sequence[str],
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.MultipleAlbumsData:

        if len(ids) > 20:
            raise ValueError("'ids' can not contain more than 20 ids.")

        parameters: typings.Parameters = {"ids": ",".join(ids)}
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/albums"), parameters=parameters, credentials=credentials)

    async def get_album_tracks(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/albums/{id}/tracks", id=_id), parameters=parameters, credentials=credentials)

    async def get_saved_albums(
        self,
        *,
        market: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.UserCredentials,
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/me/albums"), parameters=parameters, credentials=credentials)

    async def save_albums(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials,
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("PUT", "/me/albums"), parameters=parameters, credentials=credentials)

    async def remove_albums(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials,
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("DELETE", "/me/albums"), parameters=parameters, credentials=credentials)

    async def check_saved_albums(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials,
    ) -> list[bool]:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("GET", "/me/albums/contains"), parameters=parameters, credentials=credentials)

    async def get_new_releases(
        self,
        *,
        country: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.NewReleasesData:

        parameters: typings.Parameters = {}
        if country:
            parameters["country"] = country
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/browse/new-releases"), parameters=parameters, credentials=credentials)

    # ARTISTS API

    async def get_artist(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.ArtistData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/artists/{id}", id=_id), parameters=parameters, credentials=credentials)

    async def get_artists(
        self,
        ids: Sequence[str],
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.MultipleArtistsData:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/artists"), parameters=parameters, credentials=credentials)

    async def get_artist_albums(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        limit: int | None,
        offset: int | None,
        include_groups: Sequence[objects.IncludeGroup] | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset
        if include_groups:
            parameters["include_groups"] = ",".join(include_group.value for include_group in include_groups)

        return await self.request(Route("GET", "/artists/{id}/albums", id=_id), parameters=parameters, credentials=credentials)

    async def get_artist_top_tracks(
        self,
        _id: str,
        /,
        *,
        market: str,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.ArtistTopTracksData:

        parameters: typings.Parameters = {
            "market": market
        }
        return await self.request(Route("GET", "/artists/{id}/top-tracks", id=_id), parameters=parameters, credentials=credentials)

    async def get_related_artists(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.ArtistRelatedArtistsData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/artists/{id}/related-artists", id=_id), parameters=parameters, credentials=credentials)

    # SHOWS API

    async def get_show(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.ShowData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/shows/{id}", id=_id), parameters=parameters, credentials=credentials)

    async def get_shows(
        self,
        ids: Sequence[str],
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.MultipleShowsData:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/shows"), parameters=parameters, credentials=credentials)

    async def get_show_episodes(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/shows/{id}/episodes", id=_id), parameters=parameters, credentials=credentials)

    async def get_saved_shows(
        self,
        *,
        limit: int | None,
        offset: int | None,
        credentials: objects.UserCredentials,
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/me/shows"), parameters=parameters, credentials=credentials)

    async def save_shows(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials,
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("PUT", "/me/shows"), parameters=parameters, credentials=credentials)

    async def remove_shows(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials,
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("DELETE", "/me/shows"), parameters=parameters, credentials=credentials)

    async def check_saved_shows(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials,
    ) -> list[bool]:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("GET", "/me/shows/contains"), parameters=parameters, credentials=credentials)

    # EPISODE API

    async def get_episode(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.EpisodeData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/episodes/{id}", id=_id), parameters=parameters, credentials=credentials)

    async def get_episodes(
        self,
        ids: Sequence[str],
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.MultipleEpisodesData:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/episodes"), parameters=parameters, credentials=credentials)

    async def get_saved_episodes(
        self,
        *,
        market: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.UserCredentials,
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/me/episodes"), parameters=parameters, credentials=credentials)

    async def save_episodes(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("PUT", "/me/episodes"), parameters=parameters, credentials=credentials)

    async def remove_episodes(
        self,
        ids: list[str],
        /,
        *,
        credentials: objects.UserCredentials
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("DELETE", "/me/episodes"), parameters=parameters, credentials=credentials)

    async def check_saved_episodes(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials
    ) -> list[bool]:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("GET", "/me/episodes/contains"), parameters=parameters, credentials=credentials)

    # TRACKS API

    async def get_track(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.TrackData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/tracks/{id}", id=_id), parameters=parameters, credentials=credentials)

    async def get_tracks(
        self,
        ids: Sequence[str],
        *,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.MultipleTracksData:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/tracks"), parameters=parameters, credentials=credentials)

    async def get_saved_tracks(
        self,
        *,
        market: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.UserCredentials
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if market:
            parameters["market"] = market
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/me/tracks"), parameters=parameters, credentials=credentials)

    async def save_tracks(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("PUT", "/me/tracks"), parameters=parameters, credentials=credentials)

    async def remove_tracks(
        self,
        ids: list[str],
        /,
        *,
        credentials: objects.UserCredentials
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("DELETE", "/me/tracks"), parameters=parameters, credentials=credentials)

    async def check_saved_tracks(
        self,
        ids: list[str],
        /,
        *,
        credentials: objects.UserCredentials
    ) -> list[bool]:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("GET", "/me/tracks/contains"), parameters=parameters, credentials=credentials)

    async def get_track_audio_features(
        self,
        _id: str,
        /,
        *,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.AudioFeaturesData:
        return await self.request(Route("GET", "/audio-features/{id}", id=_id), credentials=credentials)

    async def get_several_tracks_audio_features(
        self,
        ids: Sequence[str],
        *,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.SeveralAudioFeaturesData:

        if len(ids) > 100:
            raise ValueError("'ids' can not contain more than 100 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(ids)
        }
        return await self.request(Route("GET", "/audio-features"), parameters=parameters, credentials=credentials)

    async def get_track_audio_analysis(
        self,
        _id: str,
        /,
        *,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> dict[str, Any]:  # TODO: create TypedDict for this
        return await self.request(Route("GET", "/audio-analysis/{id}", id=_id), credentials=credentials)

    async def get_recommendations(
        self,
        *,
        seed_artist_ids: Sequence[str] | None,
        seed_genres: Sequence[str] | None,
        seed_track_ids: Sequence[str] | None,
        limit: int | None,
        market: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None,
        **kwargs: int
    ) -> typings.RecommendationData:

        seeds = len([seed for seeds in [seed_artist_ids or [], seed_genres or [], seed_track_ids or []] for seed in seeds])
        if seeds < 1 or seeds > 5:
            raise ValueError("too many or no seed values provided. min 1, max 5.")

        parameters: typings.Parameters = {}
        if seed_artist_ids:
            parameters["seed_artists"] = ",".join(seed_artist_ids)
        if seed_genres:
            parameters["seed_genres"] = ",".join(seed_genres)
        if seed_track_ids:
            parameters["seed_tracks"] = ",".join(seed_track_ids)

        for key, value in kwargs.items():
            if key not in values.VALID_SEED_KWARGS:
                raise ValueError(f"'{key}' is not a valid kwarg.")
            parameters[key] = value

        if limit:
            utils.limit_value("limit", limit, 1, 100)
            parameters["limit"] = limit
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/recommendations"), parameters=parameters, credentials=credentials)

    # SEARCH API

    async def search(
        self,
        query: str,
        /,
        *,
        search_types: Sequence[objects.SearchType],
        include_external: bool = False,
        market: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.SearchResultData:

        parameters: typings.Parameters = {
            "q":    query.replace(" ", "+"),
            "type": ",".join(search_type.value for search_type in search_types)
        }

        if include_external:
            parameters["include_external"] = "audio"
        if market:
            parameters["market"] = market
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/search"), parameters=parameters, credentials=credentials)

    # USERS API

    async def get_current_user_profile(
        self,
        *,
        credentials: objects.UserCredentials
    ) -> typings.UserData:
        return await self.request(Route("GET", "/me"), credentials=credentials)

    async def get_current_user_top_artists(
        self,
        *,
        limit: int | None,
        offset: int | None,
        time_range: objects.TimeRange | None,
        credentials: objects.UserCredentials
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset
        if time_range:
            parameters["time_range"] = time_range.value

        return await self.request(Route("GET", "/me/top/artists"), parameters=parameters, credentials=credentials)

    async def get_current_user_top_tracks(
        self,
        *,
        limit: int | None,
        offset: int | None,
        time_range: objects.TimeRange | None,
        credentials: objects.UserCredentials
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset
        if time_range:
            parameters["time_range"] = time_range.value

        return await self.request(Route("GET", "/me/top/tracks"), parameters=parameters, credentials=credentials)

    async def get_user_profile(
        self,
        _id: str,
        /,
        *,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.UserData:
        return await self.request(Route("GET", "/users/{id}", id=_id), credentials=credentials)

    async def follow_playlist(
        self,
        _id: str,
        /,
        *,
        public: bool | None,
        credentials: objects.UserCredentials
    ) -> None:

        data: typings.Data = {}
        if public:
            data["public"] = public

        return await self.request(Route("PUT", "playlists/{id}/followers", id=_id), json=data, credentials=credentials)

    async def unfollow_playlist(
        self,
        _id: str,
        /,
        *,
        credentials: objects.UserCredentials
    ) -> None:
        return await self.request(Route("DELETE", "playlists/{id}/followers", id=_id), credentials=credentials)

    async def check_if_users_follow_playlists(
        self,
        playlist_id: str,
        /,
        *,
        user_ids: list[str],
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> None:

        if len(user_ids) > 5:
            raise ValueError("'ids' can not contain more than 5 ids.")

        parameters: typings.Parameters = {
            "ids": ",".join(user_ids)
        }
        route = Route("GET", "/playlists/{playlist_id}/followers/contains", playlist_id=playlist_id)

        return await self.request(route, parameters=parameters, credentials=credentials)

    async def get_followed_users(
        self,
    ) -> None:
        raise exceptions.SpotifyException("This operation is not yet implemented in the spotify api.")

    async def get_followed_artists(
        self,
        *,
        limit: int | None,
        after: str | None,
        credentials: objects.UserCredentials
    ) -> typings.AlternativePagingObjectData:

        parameters: typings.Parameters = {
            "type": "artist"
        }
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if after:
            parameters["after"] = after

        return await self.request(Route("GET", "/me/following"), parameters=parameters, credentials=credentials)

    async def follow_users(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "type": "user",
            "ids":  ",".join(ids)
        }

        return await self.request(Route("PUT", "/me/following"), parameters=parameters, credentials=credentials)

    async def follow_artists(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "type": "artist",
            "ids":  ",".join(ids)
        }

        return await self.request(Route("PUT", "/me/following"), parameters=parameters, credentials=credentials)

    async def unfollow_users(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "type": "user",
            "ids":  ",".join(ids)
        }

        return await self.request(Route("DELETE", "/me/following"), parameters=parameters, credentials=credentials)

    async def unfollow_artists(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials
    ) -> None:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "type": "artist",
            "ids":  ",".join(ids)
        }

        return await self.request(Route("DELETE", "/me/following"), parameters=parameters, credentials=credentials)

    async def check_followed_users(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials
    ) -> list[bool]:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "type": "user",
            "ids":  ",".join(ids)
        }

        return await self.request(Route("GET", "/me/following/contains"), parameters=parameters, credentials=credentials)

    async def check_followed_artists(
        self,
        ids: list[str],
        *,
        credentials: objects.UserCredentials
    ) -> list[bool]:

        if len(ids) > 50:
            raise ValueError("'ids' can not contain more than 50 ids.")

        parameters: typings.Parameters = {
            "type": "artist",
            "ids":  ",".join(ids)
        }

        return await self.request(Route("GET", "/me/following/contains"), parameters=parameters, credentials=credentials)

    # PLAYLISTS API

    async def get_playlist(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        fields: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.PlaylistData:

        parameters: typings.Parameters = {
            "additional_types": "track"
        }  # TODO: Support all types
        if market:
            parameters["market"] = market
        if fields:
            parameters["fields"] = fields

        return await self.request(Route("GET", "/playlists/{id}", id=_id), parameters=parameters, credentials=credentials)

    async def change_playlist_details(
        self,
        _id: str,
        /,
        *,
        name: str | None,
        public: bool | None,
        collaborative: bool | None,
        description: str | None,
        credentials: objects.UserCredentials
    ) -> None:

        if collaborative and public:
            raise ValueError("collaborative playlists can not be public.")

        data: typings.Data = {}
        if name:
            data["name"] = name
        if public:
            data["public"] = public
        if collaborative:
            data["collaborative"] = collaborative
        if description:
            data["description"] = description

        return await self.request(Route("PUT", "/playlists/{id}", id=_id), json=data, credentials=credentials)

    async def get_playlist_items(
        self,
        _id: str,
        /,
        *,
        market: str | None,
        fields: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {
            "additional_types": "track"
        }  # TODO: Support all types
        if market:
            parameters["market"] = market
        if fields:
            parameters["fields"] = fields
        if limit:
            utils.limit_value("limit", limit, 1, 100)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/playlists/{id}/tracks", id=_id), parameters=parameters, credentials=credentials)

    async def add_items_to_playlist(
        self,
        _id: str,
        /,
        *,
        uris: Sequence[str],
        position: int | None,
        credentials: objects.UserCredentials
    ) -> typings.PlaylistSnapshotId:

        if len(uris) > 100:
            raise ValueError("'uris' can not contain more than 100 URI's.")

        data: typings.Data = {
            "uris": uris
        }
        if position:
            data["position"] = position

        return await self.request(Route("POST", "/playlists/{id}/tracks", id=_id), json=data, credentials=credentials)

    async def reorder_playlist_items(
        self,
        _id: str,
        /,
        *,
        range_start: int,
        insert_before: int,
        range_length: int | None,
        snapshot_id: str | None,
        credentials: objects.UserCredentials
    ) -> typings.PlaylistSnapshotId:

        data: typings.Data = {
            "range_start":   range_start,
            "insert_before": insert_before
        }
        if range_length:
            data["range_length"] = range_length
        if snapshot_id:
            data["snapshot_id"] = snapshot_id

        return await self.request(Route("PUT", "/playlists/{id}/tracks", id=_id), json=data, credentials=credentials)

    async def replace_playlist_items(
        self,
        _id: str,
        /,
        *,
        uris: Sequence[str],
        credentials: objects.UserCredentials
    ) -> None:

        if len(uris) > 100:
            raise ValueError("'uris' can not contain more than 100 URI's.")

        data: typings.Data = {
            "uris": uris
        }
        return await self.request(Route("PUT", "/playlists/{id}/tracks", id=_id), json=data, credentials=credentials)

    async def remove_items_from_playlist(
        self,
        _id: str,
        /,
        *,
        uris: Sequence[str],
        snapshot_id: str | None,
        credentials: objects.UserCredentials
    ) -> typings.PlaylistSnapshotId:

        if len(uris) > 100:
            raise ValueError("'uris' can not contain more than 100 URI's.")

        data: typings.Data = {
            "tracks": [{"uri": uri} for uri in uris]
        }
        if snapshot_id:
            data["snapshot_id"] = snapshot_id

        return await self.request(Route("DELETE", "/playlists/{id}/tracks", id=_id), json=data, credentials=credentials)

    async def get_current_user_playlists(
        self,
        *,
        limit: int | None,
        offset: int | None,
        credentials: objects.UserCredentials
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/me/playlists"), parameters=parameters, credentials=credentials)

    async def get_user_playlists(
        self,
        _id: str,
        /,
        *,
        limit: int | None,
        offset: int | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.PagingObjectData:

        parameters: typings.Parameters = {}
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/users/{id}/playlists", id=_id), parameters=parameters, credentials=credentials)

    async def create_playlist(
        self,
        *,
        user_id: str,
        name: str,
        public: bool | None,
        collaborative: bool | None,
        description: str | None,
        credentials: objects.UserCredentials
    ) -> typings.PlaylistData:

        if collaborative and public:
            raise ValueError("collaborative playlists can not be public.")

        data: typings.Data = {
            "name": name
        }
        if public:
            data["public"] = public
        if collaborative:
            data["collaborative"] = collaborative
        if description:
            data["description"] = description

        return await self.request(Route("POST", "/users/{user_id}/playlists", user_id=user_id), json=data, credentials=credentials)

    async def get_featured_playlists(
        self,
        *,
        country: str | None,
        locale: str | None,
        timestamp: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.FeaturedPlaylistsData:

        parameters: typings.Parameters = {}
        if country:
            parameters["country"] = country
        if locale:
            parameters["locale"] = locale
        if timestamp:
            parameters["timestamp"] = timestamp
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/browse/featured-playlists"), parameters=parameters, credentials=credentials)

    async def get_category_playlists(
        self,
        _id: str,
        /,
        *,
        country: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.CategoryPlaylistsData:

        parameters: typings.Parameters = {}
        if country:
            parameters["country"] = country
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/browse/categories/{id}/playlists", id=_id), parameters=parameters, credentials=credentials)

    async def get_playlist_cover_image(
        self,
        _id: str,
        /,
        *,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> Sequence[typings.ImageData]:
        return await self.request(Route("GET", "/playlists/{id}/images", id=_id), credentials=credentials)

    async def upload_playlist_cover_image(
        self,
        _id: str,
        /,
        *,
        url: str,
        credentials: objects.UserCredentials
    ) -> None:

        if not self._session:
            self._session = aiohttp.ClientSession()

        async with self._session.get(url) as request:

            if request.status != 200:
                raise exceptions.SpotifyException("There was a problem while uploading that image.")

            image_bytes = await request.read()
            data = base64.b64encode(image_bytes).decode("utf-8")

        return await self.request(Route("PUT", "/playlists/{id}/images", id=_id), data=data, credentials=credentials)

    # CATEGORY API

    async def get_categories(
        self,
        *,
        country: str | None,
        locale: str | None,
        limit: int | None,
        offset: int | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.MultipleCategoriesData:

        parameters: typings.Parameters = {}
        if country:
            parameters["country"] = country
        if locale:
            parameters["locale"] = locale
        if limit:
            utils.limit_value("limit", limit, 1, 50)
            parameters["limit"] = limit
        if offset:
            parameters["offset"] = offset

        return await self.request(Route("GET", "/browse/categories"), parameters=parameters, credentials=credentials)

    async def get_category(
        self,
        _id: str,
        /,
        *,
        country: str | None,
        locale: str | None,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.CategoryData:

        parameters: typings.Parameters = {}
        if country:
            parameters["country"] = country
        if locale:
            parameters["locale"] = locale

        return await self.request(Route("GET", "/browse/categories/{id}", id=_id), parameters=parameters, credentials=credentials)

    # GENRE API

    async def get_available_genre_seeds(
        self,
        *,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.RecommendationGenresData:
        return await self.request(Route("GET", "/recommendations/available-genre-seeds"), credentials=credentials)

    # PLAYER API

    async def get_playback_state(
        self,
        *,
        market: str | None,
        credentials: objects.UserCredentials
    ) -> typings.PlaybackStateData:

        parameters: typings.Parameters = {
            "additional_types": "track"
        }  # TODO: Support all types
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/me/player"), parameters=parameters, credentials=credentials)

    async def transfer_playback(
        self,
        *,
        device_id: str,
        ensure_playback: bool | None,
        credentials: objects.UserCredentials
    ) -> None:

        data: typings.Data = {
            "device_ids": [device_id]
        }
        if ensure_playback:
            data["play"] = ensure_playback

        return await self.request(Route("PUT", "/me/player"), json=data, credentials=credentials)

    async def get_available_devices(
        self,
        *,
        credentials: objects.UserCredentials
    ) -> typings.MultipleDevicesData:
        return await self.request(Route("GET", "/me/player/devices"), credentials=credentials)

    async def get_currently_playing_track(
        self,
        *,
        market: str | None,
        credentials: objects.UserCredentials
    ) -> typings.CurrentlyPlayingData:

        parameters: typings.Parameters = {
            "additional_types": "track"
        }  # TODO: Support all types
        if market:
            parameters["market"] = market

        return await self.request(Route("GET", "/me/player/currently-playing"), parameters=parameters, credentials=credentials)

    async def start_playback(
        self,
        *,
        device_id: str | None,
        context_uri: str | None,
        uris: list[str] | None,
        offset: int | str | None,
        position_ms: int | None,
        credentials: objects.UserCredentials
    ) -> None:

        if context_uri and uris:
            raise ValueError("'context_uri' and 'uris' can not both be specified.")

        parameters: typings.Parameters = {}
        if device_id:
            parameters["device_id"] = device_id

        data: typings.Data = {}

        if context_uri or uris:

            if context_uri:
                data["context_uri"] = context_uri
            if uris:
                data["uris"] = uris

            if offset:
                data["offset"] = {}
                if isinstance(offset, int):
                    data["offset"]["position"] = offset
                else:
                    data["offset"]["uri"] = offset

            if position_ms:
                data["position_ms"] = position_ms

        return await self.request(Route("PUT", "/me/player/play"), parameters=parameters, json=data, credentials=credentials)

    async def resume_playback(
        self,
        *,
        device_id: str | None,
        offset: int | str | None,
        position_ms: int | None,
        credentials: objects.UserCredentials
    ) -> None:

        return await self.start_playback(
            device_id=device_id,
            context_uri=None,
            uris=None,
            offset=offset,
            position_ms=position_ms,
            credentials=credentials
        )

    async def pause_playback(
        self,
        *,
        device_id: str | None,
        credentials: objects.UserCredentials
    ) -> None:

        parameters: typings.Parameters = {}
        if device_id:
            parameters["device_id"] = device_id

        return await self.request(Route("PUT", "/me/player/pause"), parameters=parameters, credentials=credentials)

    async def skip_to_next(
        self,
        *,
        device_id: str | None,
        credentials: objects.UserCredentials
    ) -> None:

        parameters: typings.Parameters = {}
        if device_id:
            parameters["device_id"] = device_id

        return await self.request(Route("POST", "/me/player/next"), parameters=parameters, credentials=credentials)

    async def skip_to_previous(
        self,
        *,
        device_id: str | None,
        credentials: objects.UserCredentials
    ) -> None:

        parameters: typings.Parameters = {}
        if device_id:
            parameters["device_id"] = device_id

        return await self.request(Route("POST", "/me/player/previous"), parameters=parameters, credentials=credentials)

    async def seek_to_position(
        self,
        *,
        position_ms: int,
        device_id: str | None,
        credentials: objects.UserCredentials
    ) -> None:

        parameters: typings.Parameters = {
            "position_ms": position_ms
        }
        if device_id:
            parameters["device_id"] = device_id

        return await self.request(Route("PUT", "/me/player/seek"), parameters=parameters, credentials=credentials)

    async def set_repeat_mode(
        self,
        *,
        repeat_mode: objects.RepeatMode,
        device_id: str | None,
        credentials: objects.UserCredentials
    ) -> None:

        parameters: typings.Parameters = {
            "state": repeat_mode.value
        }
        if device_id:
            parameters["device_id"] = device_id

        return await self.request(Route("PUT", "/me/player/repeat"), parameters=parameters, credentials=credentials)

    async def set_playback_volume(
        self,
        *,
        volume_percent: int,
        device_id: str | None,
        credentials: objects.UserCredentials
    ) -> None:

        utils.limit_value("volume_percent", volume_percent, 0, 100)

        parameters: typings.Parameters = {
            "volume_percent": volume_percent
        }
        if device_id:
            parameters["device_id"] = device_id

        return await self.request(Route("PUT", "/me/player/volume"), parameters=parameters, credentials=credentials)

    async def toggle_playback_shuffle(
        self,
        *,
        state: bool,
        device_id: str | None,
        credentials: objects.UserCredentials
    ) -> None:

        parameters: typings.Parameters = {
            "state": "true" if state else "false"
        }
        if device_id:
            parameters["device_id"] = device_id

        return await self.request(Route("PUT", "/me/player/shuffle"), parameters=parameters, credentials=credentials)

    async def get_recently_played_tracks(
        self,
        *,
        limit: int | None,
        before: int | None,
        after: int | None,
        credentials: objects.UserCredentials
    ) -> typings.AlternativePagingObjectData:

        if before and after:
            raise ValueError("'before' and 'after' can not both be specified.")

        parameters: typings.Parameters = {}
        if limit:
            parameters["limit"] = limit
        if before:
            parameters["before"] = before
        if after:
            parameters["after"] = after

        return await self.request(Route("GET", "/me/player/recently-played"), parameters=parameters, credentials=credentials)

    async def add_item_to_playback_queue(
        self,
        *,
        uri: str,
        device_id: str | None,
        credentials: objects.UserCredentials
    ) -> None:

        parameters: typings.Parameters = {
            "uri": uri
        }
        if device_id:
            parameters["device_id"] = device_id

        return await self.request(Route("POST", "/me/player/queue"), parameters=parameters, credentials=credentials)

    # MARKETS API

    async def get_available_markets(
        self,
        *,
        credentials: objects.ClientCredentials | objects.UserCredentials | None = None
    ) -> typings.AvailableMarketsData:
        return await self.request(Route("GET", "/markets"), credentials=credentials)



# Future
from __future__ import annotations

# Standard Library
from typing import Any, Literal

# My stuff
from spotipy.typings.objects import *


HTTPMethod = Literal["GET", "POST", "DELETE", "PATCH", "PUT"]
Parameters = dict[str, Any]
Data = dict[str, Any]



# Future
from __future__ import annotations

# Standard Library
import math
from collections.abc import Sequence
from typing import TYPE_CHECKING, TypeVar

# Packages
import aiohttp

# My stuff
from spotipy import http, objects, utils


if TYPE_CHECKING:

    # My stuff
    from spotipy.typings import Credentials, OptionalCredentials


__all__ = (
    "Client",
)


ID = TypeVar("ID", bound=str)


class Client:

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        session: aiohttp.ClientSession | None = None
    ) -> None:

        self._client_id: str = client_id
        self._client_secret: str = client_secret
        self._session: aiohttp.ClientSession | None = session

        self.http: http.HTTPClient = http.HTTPClient(
            client_id=self._client_id,
            client_secret=self._client_secret,
            session=self._session
        )

    def __repr__(self) -> str:
        return "<spotipy.Client>"

    #

    async def close(self) -> None:

        if not self.http._session:
            return

        await self.http.close()

    # ALBUMS API

    async def get_album(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> objects.Album:

        response = await self.http.get_album(_id, market=market, credentials=credentials)
        return objects.Album(response)

    async def get_albums(
        self,
        ids: Sequence[ID],
        *,
        market: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> dict[ID, objects.Album | None]:

        response = await self.http.get_albums(ids=ids, market=market, credentials=credentials)
        return dict(zip(ids, [objects.Album(data) if data else None for data in response["albums"]]))

    async def get_album_tracks(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.SimpleTrack]:

        response = await self.http.get_album_tracks(_id, market=market, limit=limit, offset=offset, credentials=credentials)
        return [objects.SimpleTrack(data) for data in objects.PagingObject(response).items]

    async def get_all_album_tracks(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.SimpleTrack]:

        response = await self.http.get_album_tracks(_id, market=market, limit=50, offset=0, credentials=credentials)
        paging = objects.PagingObject(response)

        tracks = [objects.SimpleTrack(data) for data in paging.items]

        if paging.total <= 50:  # There are 50 or fewer tracks, and we already have them so just return them
            return tracks

        for _ in range(1, math.ceil(paging.total / 50)):
            response = await self.http.get_album_tracks(_id, market=market, limit=50, offset=_ * 50, credentials=credentials)
            tracks.extend([objects.SimpleTrack(data) for data in objects.PagingObject(response).items])

        return tracks

    async def get_full_album(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> objects.Album:

        album = await self.get_album(_id, market=market, credentials=credentials)

        if album._tracks_paging.total <= 50:  # The album has 50 or fewer tracks already, so we can just return it now.
            return album

        for _ in range(2, math.ceil(album._tracks_paging.total / 50)):
            response = await self.http.get_album_tracks(_id, market=market, limit=50, offset=_ * 50, credentials=credentials)
            album.tracks.extend([objects.SimpleTrack(data) for data in objects.PagingObject(response).items])

        return album

    async def get_saved_albums(self) -> ...:
        raise NotImplementedError

    async def save_albums(self) -> ...:
        raise NotImplementedError

    async def remove_albums(self) -> ...:
        raise NotImplementedError

    async def check_saved_albums(self) -> ...:
        raise NotImplementedError

    async def get_new_releases(
        self,
        *,
        country: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.SimpleAlbum]:

        response = await self.http.get_new_releases(country=country, limit=limit, offset=offset, credentials=credentials)
        return [objects.SimpleAlbum(data) for data in objects.PagingObject(response["albums"]).items]

    # ARTISTS API

    async def get_artist(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> objects.Artist:

        response = await self.http.get_artist(_id, market=market, credentials=credentials)
        return objects.Artist(response)

    async def get_artists(
        self,
        ids: Sequence[ID],
        *,
        market: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> dict[ID, objects.Artist | None]:

        response = await self.http.get_artists(ids=ids, market=market, credentials=credentials)
        return dict(zip(ids, [objects.Artist(data) if data else None for data in response["artists"]]))

    async def get_artist_albums(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        include_groups: Sequence[objects.IncludeGroup] | None = utils.MISSING,
        limit: int | None = None,
        offset: int | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.SimpleAlbum]:

        if include_groups == utils.MISSING:
            include_groups = [objects.IncludeGroup.ALBUM]

        response = await self.http.get_artist_albums(_id, market=market, include_groups=include_groups, limit=limit, offset=offset, credentials=credentials)
        return [objects.SimpleAlbum(data) for data in objects.PagingObject(response).items]

    async def get_all_artist_albums(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        include_groups: Sequence[objects.IncludeGroup] | None = utils.MISSING,
        credentials: OptionalCredentials = None,
    ) -> list[objects.SimpleAlbum]:

        if include_groups == utils.MISSING:
            include_groups = [objects.IncludeGroup.ALBUM]

        response = await self.http.get_artist_albums(_id, market=market, include_groups=include_groups, limit=50, offset=0, credentials=credentials)
        paging = objects.PagingObject(response)

        albums = [objects.SimpleAlbum(data) for data in paging.items]

        if paging.total <= 50:  # There are 50 or fewer tracks, and we already have them so just return them
            return albums

        for _ in range(1, math.ceil(paging.total / 50)):
            response = await self.http.get_artist_albums(_id, market=market, include_groups=include_groups, limit=50, offset=_ * 50, credentials=credentials)
            albums.extend([objects.SimpleAlbum(data) for data in objects.PagingObject(response).items])

        return albums

    async def get_artist_top_tracks(
        self,
        _id: str,
        /,
        *,
        market: str = "GB",
        credentials: OptionalCredentials = None,
    ) -> list[objects.Track]:

        response = await self.http.get_artist_top_tracks(_id, market=market, credentials=credentials)
        return [objects.Track(data) for data in response["tracks"]]

    async def get_related_artists(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.Artist]:

        response = await self.http.get_related_artists(_id, market=market, credentials=credentials)
        return [objects.Artist(data) for data in response["artists"]]

    # SHOWS API

    async def get_show(
        self,
        _id: str,
        /,
        *,
        market: str | None = "GB",
        credentials: OptionalCredentials = None,
    ) -> objects.Show:

        response = await self.http.get_show(_id, market=market, credentials=credentials)
        return objects.Show(response)

    async def get_shows(
        self,
        ids: Sequence[ID],
        *,
        market: str | None = "GB",
        credentials: OptionalCredentials = None,
    ) -> dict[ID, objects.Show | None]:

        response = await self.http.get_shows(ids, market=market, credentials=credentials)
        return dict(zip(ids, [objects.Show(data) if data else None for data in response["shows"]]))

    async def get_show_episodes(
        self,
        _id: str,
        /,
        *,
        market: str | None = "GB",
        limit: int | None = None,
        offset: int | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.SimpleEpisode]:

        response = await self.http.get_show_episodes(_id, market=market, limit=limit, offset=offset, credentials=credentials)
        return [objects.SimpleEpisode(data) for data in objects.PagingObject(response).items]

    async def get_all_show_episodes(
        self,
        _id: str,
        /,
        *,
        market: str | None = "GB",
        credentials: OptionalCredentials = None,
    ) -> list[objects.SimpleEpisode]:

        response = await self.http.get_show_episodes(_id, market=market, limit=50, offset=0, credentials=credentials)
        paging = objects.PagingObject(response)

        episodes = [objects.SimpleEpisode(data) for data in paging.items]

        if paging.total <= 50:  # There are 50 or fewer episodes, and we already have them so just return them
            return episodes

        for _ in range(1, math.ceil(paging.total / 50)):
            response = await self.http.get_show_episodes(_id, market=market, limit=50, offset=_ * 50, credentials=credentials)
            episodes.extend([objects.SimpleEpisode(data) for data in objects.PagingObject(response).items])

        return episodes

    async def get_saved_shows(self) -> ...:
        raise NotImplementedError

    async def save_shows(self) -> ...:
        raise NotImplementedError

    async def remove_shows(self) -> ...:
        raise NotImplementedError

    async def check_saved_shows(self) -> ...:
        raise NotImplementedError

    # EPISODE API

    async def get_episode(
        self,
        _id: str,
        /,
        *,
        market: str | None = "GB",
        credentials: OptionalCredentials = None,
    ) -> objects.Episode:

        response = await self.http.get_episode(_id, market=market, credentials=credentials)
        return objects.Episode(response)

    async def get_episodes(
        self,
        ids: Sequence[ID],
        *,
        market: str | None = "GB",
        credentials: OptionalCredentials = None,
    ) -> dict[ID, objects.Episode | None]:

        response = await self.http.get_episodes(ids=ids, market=market, credentials=credentials)
        return dict(zip(ids, [objects.Episode(data) if data else None for data in response["episodes"]]))

    async def get_saved_episodes(self) -> ...:
        raise NotImplementedError

    async def save_episodes(self) -> ...:
        raise NotImplementedError

    async def remove_episodes(self) -> ...:
        raise NotImplementedError

    async def check_saved_episodes(self) -> ...:
        raise NotImplementedError

    # TRACKS API

    async def get_track(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> objects.Track:

        response = await self.http.get_track(_id, market=market, credentials=credentials)
        return objects.Track(response)

    async def get_tracks(
        self,
        ids: Sequence[ID],
        *,
        market: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> dict[ID, objects.Track | None]:

        response = await self.http.get_tracks(ids=ids, market=market, credentials=credentials)
        return dict(zip(ids, [objects.Track(data) if data else None for data in response["tracks"]]))

    async def get_saved_tracks(self) -> ...:
        raise NotImplementedError

    async def save_tracks(self) -> ...:
        raise NotImplementedError

    async def remove_tracks(self) -> ...:
        raise NotImplementedError

    async def check_saved_tracks(self) -> ...:
        raise NotImplementedError

    async def get_several_tracks_audio_features(
        self,
        ids: Sequence[ID],
        *,
        credentials: OptionalCredentials = None,
    ) -> dict[ID, objects.AudioFeatures | None]:

        response = await self.http.get_several_tracks_audio_features(ids, credentials=credentials)
        return dict(zip(ids, [objects.AudioFeatures(data) if data else None for data in response["audio_features"]]))

    async def get_track_audio_features(
        self,
        _id: str,
        /,
        *,
        credentials: OptionalCredentials = None,
    ) -> objects.AudioFeatures:

        response = await self.http.get_track_audio_features(_id, credentials=credentials)
        return objects.AudioFeatures(response)

    async def get_track_audio_analysis(self) -> ...:
        raise NotImplementedError

    async def get_recommendations(
        self,
        *,
        seed_artist_ids: Sequence[str] | None = None,
        seed_genres: Sequence[str] | None = None,
        seed_track_ids: Sequence[str] | None = None,
        limit: int | None = None,
        market: str | None = None,
        credentials: OptionalCredentials = None,
        **kwargs: int
    ) -> objects.Recommendation:

        response = await self.http.get_recommendations(
            seed_artist_ids=seed_artist_ids,
            seed_genres=seed_genres,
            seed_track_ids=seed_track_ids,
            limit=limit,
            market=market,
            credentials=credentials,
            **kwargs
        )
        return objects.Recommendation(response)

    # SEARCH API

    async def search(
        self,
        query: str,
        /,
        *,
        search_types: Sequence[objects.SearchType] | None = None,
        market: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        include_external: bool = False,
        credentials: OptionalCredentials = None,
    ) -> objects.SearchResult:

        response = await self.http.search(
            query,
            search_types=search_types,
            market=market,
            limit=limit,
            offset=offset,
            include_external=include_external,
            credentials=credentials
        )
        return objects.SearchResult(response)

    # USERS API

    async def get_current_user_profile(
        self,
        *,
        credentials: Credentials,
    ) -> objects.User:

        response = await self.http.get_current_user_profile(credentials=credentials)
        return objects.User(response)

    async def get_current_users_top_artists(
        self,
        *,
        time_range: objects.TimeRange | None,
        limit: int | None,
        offset: int | None,
        credentials: Credentials,
    ) -> list[objects.Artist]:

        response = await self.http.get_current_user_top_artists(time_range=time_range, limit=limit, offset=offset, credentials=credentials)
        return [objects.Artist(data) for data in objects.PagingObject(response).items]

    async def get_current_users_top_tracks(
        self,
        *,
        time_range: objects.TimeRange | None = None,
        limit: int | None = None,
        offset: int | None = None,
        credentials: Credentials,
    ) -> list[objects.Track]:

        response = await self.http.get_current_user_top_tracks(time_range=time_range, limit=limit, offset=offset, credentials=credentials)
        return [objects.Track(data) for data in objects.PagingObject(response).items]

    async def get_user_profile(
        self,
        _id: str,
        /,
        *,
        credentials: Credentials,
    ) -> objects.User:

        response = await self.http.get_user_profile(_id, credentials=credentials)
        return objects.User(response)

    async def follow_playlist(self) -> ...:
        raise NotImplementedError

    async def unfollow_playlist(self) -> ...:
        raise NotImplementedError

    async def get_followed_artists(self) -> ...:
        raise NotImplementedError

    async def get_followed_users(self) -> ...:
        raise NotImplementedError

    async def follow_artists(self) -> ...:
        raise NotImplementedError

    async def follow_users(self) -> ...:
        raise NotImplementedError

    async def unfollow_artists(self) -> ...:
        raise NotImplementedError

    async def unfollow_users(self) -> ...:
        raise NotImplementedError

    async def check_followed_artists(self) -> ...:
        raise NotImplementedError

    async def check_followed_users(self) -> ...:
        raise NotImplementedError

    async def check_playlist_followers(self) -> ...:
        raise NotImplementedError

    # PLAYLISTS API

    async def get_playlist(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        fields: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> objects.Playlist:

        response = await self.http.get_playlist(_id, market=market, fields=fields, credentials=credentials)
        return objects.Playlist(response)

    async def change_playlist_details(self) -> ...:
        raise NotImplementedError

    async def get_playlist_items(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        fields: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.PlaylistTrack]:

        response = await self.http.get_playlist_items(_id, market=market, fields=fields, limit=limit, offset=offset, credentials=credentials)
        return [objects.PlaylistTrack(data) for data in objects.PagingObject(response).items]

    async def get_all_playlist_items(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        fields: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.PlaylistTrack]:

        response = await self.http.get_playlist_items(_id, market=market, fields=fields, limit=100, offset=0, credentials=credentials)
        paging = objects.PagingObject(response)

        items = [objects.PlaylistTrack(data) for data in paging.items]

        if paging.total <= 100:  # There are 50 or fewer tracks, and we already have them so just return them
            return items

        for _ in range(1, math.ceil(paging.total / 100)):
            response = await self.http.get_playlist_items(_id, market=market, fields=fields, limit=100, offset=_ * 100, credentials=credentials)
            items.extend([objects.PlaylistTrack(data) for data in objects.PagingObject(response).items])

        return items

    async def get_full_playlist(
        self,
        _id: str,
        /,
        *,
        market: str | None = None,
        fields: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> objects.Playlist:

        playlist = await self.get_playlist(_id, market=market, fields=fields, credentials=credentials)

        if playlist._tracks_paging.total <= 100:  # The playlist has 100 or fewer tracks already, so we can just return it now.
            return playlist

        for _ in range(1, math.ceil(playlist._tracks_paging.total / 100)):
            response = await self.http.get_playlist_items(_id, market=market, fields=fields, limit=100, offset=_ * 100, credentials=credentials)
            playlist.tracks.extend([objects.PlaylistTrack(data) for data in objects.PagingObject(response).items])

        return playlist

    async def add_items_to_playlist(self) -> ...:
        raise NotImplementedError

    async def reorder_playlist_items(self) -> ...:
        raise NotImplementedError

    async def replace_playlist_items(self) -> ...:
        raise NotImplementedError

    async def remove_items_from_playlist(self) -> ...:
        raise NotImplementedError

    async def get_current_user_playlists(
        self,
        *,
        limit: int | None = None,
        offset: int | None = None,
        credentials: Credentials,
    ) -> list[objects.SimplePlaylist]:

        response = await self.http.get_current_user_playlists(limit=limit, offset=offset, credentials=credentials)
        return [objects.SimplePlaylist(data) for data in objects.PagingObject(response).items]

    async def get_all_current_user_playlists(
        self,
        *,
        credentials: Credentials,
    ) -> list[objects.SimplePlaylist]:

        response = await self.http.get_current_user_playlists(limit=50, offset=0, credentials=credentials)
        paging = objects.PagingObject(response)

        playlists = [objects.SimplePlaylist(data) for data in paging.items]

        if paging.total <= 50:  # There are 50 or fewer playlists, and we already have them so just return them
            return playlists

        for _ in range(1, math.ceil(paging.total / 50)):
            response = await self.http.get_current_user_playlists(limit=50, offset=_ * 50, credentials=credentials)
            playlists.extend([objects.SimplePlaylist(data) for data in objects.PagingObject(response).items])

        return playlists

    async def get_user_playlists(
        self,
        _id: str,
        /,
        *,
        limit: int | None = None,
        offset: int | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.SimplePlaylist]:

        response = await self.http.get_user_playlists(_id, limit=limit, offset=offset, credentials=credentials)
        return [objects.SimplePlaylist(data) for data in objects.PagingObject(response).items]

    async def get_all_user_playlists(
        self,
        _id: str,
        /,
        *,
        credentials: OptionalCredentials = None,
    ) -> list[objects.SimplePlaylist]:

        response = await self.http.get_user_playlists(_id, limit=50, offset=0, credentials=credentials)
        paging = objects.PagingObject(response)

        playlists = [objects.SimplePlaylist(data) for data in paging.items]

        if paging.total <= 50:  # There are 50 or fewer playlists, and we already have them so just return them
            return playlists

        for _ in range(1, math.ceil(paging.total / 50)):
            response = await self.http.get_user_playlists(_id, limit=50, offset=_ * 50, credentials=credentials)
            playlists.extend([objects.SimplePlaylist(data) for data in objects.PagingObject(response).items])

        return playlists

    async def create_playlist(self) -> ...:
        raise NotImplementedError

    async def get_featured_playlists(
        self,
        *,
        country: str | None = None,
        locale: str | None = None,
        timestamp: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        credentials: OptionalCredentials = None,
    ) -> tuple[str, list[objects.SimplePlaylist]]:

        response = await self.http.get_featured_playlists(
            country=country,
            locale=locale,
            timestamp=timestamp,
            limit=limit,
            offset=offset,
            credentials=credentials
        )
        return response["message"], [objects.SimplePlaylist(data) for data in objects.PagingObject(response["playlists"]).items]

    async def get_category_playlists(
        self,
        _id: str,
        /,
        *,
        country: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.SimplePlaylist]:

        response = await self.http.get_category_playlists(_id, country=country, limit=limit, offset=offset, credentials=credentials)
        return [objects.SimplePlaylist(data) for data in objects.PagingObject(response["playlists"]).items]

    async def get_playlist_cover_image(
        self,
        _id: str,
        /,
        *,
        credentials: OptionalCredentials = None
    ) -> list[objects.Image]:

        response = await self.http.get_playlist_cover_image(_id, credentials=credentials)
        return [objects.Image(data) for data in response]

    async def upload_playlist_cover_image(self) -> ...:
        raise NotImplementedError

    # CATEGORY API

    async def get_categories(
        self,
        *,
        country: str | None = None,
        locale: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        credentials: OptionalCredentials = None,
    ) -> list[objects.Category]:

        response = await self.http.get_categories(country=country, locale=locale, limit=limit, offset=offset, credentials=credentials)
        return [objects.Category(data) for data in objects.PagingObject(response["categories"]).items]

    async def get_category(
        self,
        _id: str,
        /,
        *,
        country: str | None = None,
        locale: str | None = None,
        credentials: OptionalCredentials = None,
    ) -> objects.Category:

        response = await self.http.get_category(_id, country=country, locale=locale, credentials=credentials)
        return objects.Category(response)

    # GENRE API

    async def get_available_genre_seeds(
        self,
        *,
        credentials: OptionalCredentials = None,
    ) -> list[str]:

        response = await self.http.get_available_genre_seeds(credentials=credentials)
        return response["genres"]

    # PLAYER API

    async def get_playback_state(self) -> ...:
        raise NotImplementedError

    async def transfer_playback(self) -> ...:
        raise NotImplementedError

    async def get_available_devices(self) -> ...:
        raise NotImplementedError

    async def get_currently_playing_track(self) -> ...:
        raise NotImplementedError

    async def start_playback(self) -> ...:
        raise NotImplementedError

    async def resume_playback(self) -> ...:
        raise NotImplementedError

    async def pause_playback(self) -> ...:
        raise NotImplementedError

    async def skip_to_next(self) -> ...:
        raise NotImplementedError

    async def skip_to_previous(self) -> ...:
        raise NotImplementedError

    async def seek_to_position(self) -> ...:
        raise NotImplementedError

    async def set_repeat_mode(self) -> ...:
        raise NotImplementedError

    async def set_playback_volume(self) -> ...:
        raise NotImplementedError

    async def toggle_playback_shuffle(self) -> ...:
        raise NotImplementedError

    async def get_recently_played_tracks(self) -> ...:
        raise NotImplementedError

    async def add_item_to_playback_queue(self) -> ...:
        raise NotImplementedError

    # MARKETS API

    async def get_available_markets(
        self,
        *,
        credentials: OptionalCredentials = None,
    ) -> list[str]:

        response = await self.http.get_available_markets(credentials=credentials)
        return response["markets"]



# Future
from __future__ import annotations

# My stuff
from spotipy.typings.objects import ActionsData


__all__ = (
    "Actions",
)


class Actions:

    def __init__(self, data: ActionsData) -> None:
        self.interrupting_playback = data.get("interrupting_playback", False)
        self.pausing = data.get("pausing", False)
        self.resuming = data.get("resuming", False)
        self.seeking = data.get("seeking", False)
        self.skipping_next = data.get("skipping_next", False)
        self.skipping_previous = data.get("skipping_prev", False)
        self.toggling_repeat_context = data.get("toggling_repeat_context", False)
        self.toggling_repeat_track = data.get("toggling_repeat_track", False)
        self.toggling_shuffle = data.get("toggling_shuffle", False)
        self.transferring_playback = data.get("transferring_playback", False)

    def __repr__(self) -> str:
        return "<spotipy.Disallows>"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spotipy', 'spotipy.objects', 'spotipy.typings']

package_data = \
{'': ['*']}

modules = \
['py']
install_requires = \
['aiohttp>=3.6.0,<3.8.0']

extras_require = \
{'docs': ['sphinx>=4.3.1,<5.0.0',
          'sphinxcontrib-trio>=1.1.2,<2.0.0',
          'sphinx-copybutton>=0.4.0,<0.5.0',
          'sphinxext-opengraph>=0.5.0,<0.6.0',
          'furo>=2021.11.23,<2022.0.0'],
 'speed': ['orjson>=3.6.4,<4.0.0']}

setup_kwargs = {
    'name': 'spoti.py',
    'version': '0.1.1',
    'description': 'An async wrapper for the Spotify Web API.',
    'long_description': '# spoti.py\nAn async wrapper for the [Spotify API](https://developer.spotify.com/documentation/web-api/) written in Python.\n\n## Support\nSupport for spoti.py can be found in my [Discord](https://discord.com/invite/w9f6NkQbde) or by opening a discussion/issue on the [GitHub repo](https://github.com/Axelancerr/spoti.py)\n\n## Installation\n\nspoti.py is not available on PyPi yet, but it can be installed from GitHub instead:\n```shell\npip install -U git+https://github.com/Axelancerr/spoti.py\n```\n\nTo add spoti.py to a `requirements.txt` file, you can simply add it as: \n```text\ngit+https://github.com/Axelancerr/spoti.py\n```\n',
    'author': 'Axel',
    'author_email': 'axelancerr@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Axelware/spoti.py',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10.0,<4.0',
}


setup(**setup_kwargs)



# Future
from __future__ import annotations

# Standard Library
import json
from typing import Any, Literal

# Packages
import aiohttp


__all__ = (
    "to_json",
    "from_json",
    "limit_value",
    "json_or_text",
    "to_json",
    "MISSING",
)

to_json = json.dumps
from_json = json.loads


async def json_or_text(response: aiohttp.ClientResponse) -> dict[str, Any] | str:

    text = await response.text(encoding="utf-8")

    try:
        if response.headers["content-type"] in ["application/json", "application/json; charset=utf-8"]:
            return from_json(text)
    except KeyError:
        pass

    return text


def limit_value(name: str, value: int, minimum: int, maximum: int) -> None:

    if value < minimum or value > maximum:
        raise ValueError(f"'{name}' must be more than {minimum} and less than {maximum}")


class _MissingSentinel:

    def __eq__(self, other: Any) -> Literal[False]:
        return False

    def __bool__(self) -> Literal[False]:
        return False

    def __repr__(self) -> str:
        return "..."


MISSING: Any = _MissingSentinel()



# Future
from __future__ import annotations

# Standard Library
import enum


__all__ = (
    "IncludeGroup",
    "Key",
    "Mode",
    "SearchType",
    "CopyrightType",
    "RepeatMode",
    "TimeRange"
)


class IncludeGroup(enum.Enum):
    ALBUM = "album"
    SINGLE = "single"
    APPEARS_ON = "appears_on"
    COMPILATION = "compilation"
    ALL = "album,single,appears_on,compilation"


class Key(enum.Enum):
    C = 0
    C_SHARP = 1
    D = 2
    D_SHARP = 3
    E = 4
    F = 5
    F_SHARP = 6
    G = 7
    G_SHARP = 8
    A = 9
    A_SHARP = 10
    B = 11


class Mode(enum.Enum):
    MAJOR = 1
    MINOR = 0


class SearchType(enum.Enum):
    ALBUM = "album"
    ARTIST = "artist"
    PLAYLIST = "playlist"
    TRACK = "track"
    SHOW = "show"
    EPISODE = "episode"
    ALL = "album,artist,playlist,track,show,episode"


class CopyrightType(enum.Enum):
    NORMAL = "C"
    PERFORMANCE = "P"


class RepeatMode(enum.Enum):
    TRACK = "track"
    CONTEXT = "context"
    OFF = "off"


class TimeRange(enum.Enum):
    LONG_TERM = "long_term"
    MEDIUM_TERM = "medium_term"
    SHORT_TERM = "short_term"



# Future
from __future__ import annotations

# Standard Library
from typing import Any, Optional, TypedDict


__all__ = (
    "BaseObjectData",
    "PagingObjectData",
    "AlternativePagingObjectData",
    "ImageData",
    "FollowersData",
    "CopyrightData",
    "ExternalUrlsData",
    "ExternalIdsData",
    "AlbumRestrictionData",
    "SimpleAlbumData",
    "AlbumData",
    "MultipleAlbumsData",
    "NewReleasesData",
    "SimpleArtistData",
    "ArtistData",
    "MultipleArtistsData",
    "ArtistTopTracksData",
    "ArtistRelatedArtistsData",
    "ShowData",
    "MultipleShowsData",
    "EpisodeRestrictionData",
    "EpisodeResumePointData",
    "SimpleEpisodeData",
    "EpisodeData",
    "MultipleEpisodesData",
    "TrackRestrictionData",
    "SimpleTrackData",
    "TrackData",
    "MultipleTracksData",
    "AudioFeaturesData",
    "SeveralAudioFeaturesData",
    "RecommendationSeedData",
    "RecommendationData",
    "SearchResultData",
    "ExplicitContentSettingsData",
    "UserData",
    "PlaylistSnapshotId",
    "PlaylistTrackData",
    "PlaylistTrackRefData",
    "SimplePlaylistData",
    "PlaylistData",
    "FeaturedPlaylistsData",
    "CategoryPlaylistsData",
    "CategoryData",
    "MultipleCategoriesData",
    "RecommendationGenresData",
    "DeviceData",
    "MultipleDevicesData",
    "ActionsData",
    "ContextData",
    "PlaybackStateData",
    "CurrentlyPlayingData",
    "AvailableMarketsData",
    "ClientCredentialsData",
    "UserCredentialsData",
)


# COMMON

class BaseObjectData(TypedDict):
    href: str
    id: str
    name: str
    type: str
    uri: str


class PagingObjectData(TypedDict):
    href: str
    items: list[Any]  # type: ignore
    limit: int
    next: Optional[str | None]
    offset: int
    previous: Optional[str | None]
    total: int


class AlternativePagingObjectData(TypedDict):
    href: str
    items: list[Any]  # type: ignore
    limit: int
    next: Optional[str | None]
    before: str
    after: str


class ImageData(TypedDict):
    url: str
    width: int
    height: int


class FollowersData(TypedDict):
    href: Optional[str]
    total: int


class CopyrightData(TypedDict):
    text: str
    type: str


ExternalUrlsData = dict[str, Any]
ExternalIdsData = dict[str, Any]


# ALBUMS API

class AlbumRestrictionData(TypedDict):
    reason: str


class SimpleAlbumData(BaseObjectData):
    album_type: str
    artists: list[SimpleArtistData]
    available_markets: list[str]
    external_urls: ExternalUrlsData
    images: list[ImageData]
    release_date: str
    release_date_precision: str
    restrictions: AlbumRestrictionData
    total_tracks: int


class AlbumData(BaseObjectData):
    album_type: str
    artists: list[SimpleArtistData]
    available_markets: list[str]
    copyrights: list[CopyrightData]
    external_ids: ExternalIdsData
    external_urls: ExternalUrlsData
    genres: list[str]
    images: list[ImageData]
    label: str
    popularity: int
    release_date: str
    release_date_precision: str
    restrictions: AlbumRestrictionData
    total_tracks: int
    tracks: PagingObjectData


class MultipleAlbumsData(TypedDict):
    albums: list[AlbumData | None]


class NewReleasesData(TypedDict):
    albums: PagingObjectData


# ARTISTS API

class SimpleArtistData(BaseObjectData):
    external_urls: ExternalUrlsData


class ArtistData(SimpleArtistData):
    followers: FollowersData
    genres: list[str]
    images: list[ImageData]
    popularity: int


class MultipleArtistsData(TypedDict):
    artists: list[ArtistData | None]


class ArtistTopTracksData(TypedDict):
    tracks: list[TrackData]


class ArtistRelatedArtistsData(TypedDict):
    artists: list[ArtistData]


# SHOWS API

class ShowData(BaseObjectData):
    available_markets: list[str]
    copyrights: list[CopyrightData]
    description: str
    explicit: bool
    external_urls: ExternalUrlsData
    html_description: str
    images: list[ImageData]
    is_externally_hosted: bool
    languages: list[str]
    media_type: str
    publisher: str
    total_episodes: int


class MultipleShowsData(TypedDict):
    shows: list[ShowData | None]


# EPISODE API

class EpisodeRestrictionData(TypedDict):
    reason: str


class EpisodeResumePointData(TypedDict):
    fully_played: bool
    resume_position_ms: int


class SimpleEpisodeData(BaseObjectData):
    audio_preview_url: Optional[str]
    description: str
    duration_ms: int
    explicit: bool
    external_urls: ExternalUrlsData
    html_description: str
    images: list[ImageData]
    is_externally_hosted: bool
    is_playable: bool
    languages: list[str]
    release_date: str
    release_date_precision: str
    restrictions: EpisodeRestrictionData
    resume_point: EpisodeResumePointData


class EpisodeData(BaseObjectData):
    audio_preview_url: Optional[str]
    description: str
    duration_ms: int
    explicit: bool
    external_urls: ExternalUrlsData
    html_description: str
    images: list[ImageData]
    is_externally_hosted: bool
    is_playable: bool
    languages: list[str]
    release_date: str
    release_date_precision: str
    restrictions: EpisodeRestrictionData
    resume_point: EpisodeResumePointData
    show: ShowData


class MultipleEpisodesData(TypedDict):
    episodes: list[EpisodeData | None]


# TRACKS API

class TrackRestrictionData(TypedDict):
    reason: str


class SimpleTrackData(BaseObjectData):
    artists: list[SimpleArtistData]
    available_markets: list[str]
    disc_number: int
    duration_ms: int
    explicit: bool
    external_urls: ExternalUrlsData
    is_local: bool
    is_playable: bool
    #  linked_from: LinkedTrackData
    preview_url: str
    restrictions: TrackRestrictionData
    track_number: int


class TrackData(BaseObjectData):
    album: SimpleAlbumData
    artists: list[SimpleArtistData]
    available_markets: list[str]
    disc_number: int
    duration_ms: int
    explicit: bool
    external_ids: ExternalIdsData
    external_urls: ExternalUrlsData
    is_local: bool
    is_playable: bool
    #  linked_from: LinkedTrackData
    popularity: int
    preview_url: str
    restrictions: TrackRestrictionData
    track_number: int


class MultipleTracksData(TypedDict):
    tracks: list[TrackData | None]


class AudioFeaturesData(TypedDict):
    acousticness: float
    analysis_url: str
    danceability: float
    duration_ms: int
    energy: float
    id: str
    instrumentalness: float
    key: int
    liveness: float
    loudness: float
    mode: int
    speechiness: float
    tempo: float
    time_signature: int
    track_href: str
    type: str
    uri: str
    valence: float


class SeveralAudioFeaturesData(TypedDict):
    audio_features: list[AudioFeaturesData | None]


class RecommendationSeedData(TypedDict):
    initialPoolSize: int
    afterFilteringSize: int
    afterRelinkingSize: int
    id: str
    type: str
    href: str


class RecommendationData(TypedDict):
    tracks: list[TrackData]
    seeds: list[RecommendationSeedData]


# SEARCH API

class SearchResultData(TypedDict):
    albums: PagingObjectData
    artists: PagingObjectData
    tracks: PagingObjectData
    playlists: PagingObjectData
    shows: PagingObjectData
    episodes: PagingObjectData


# USERS API

class ExplicitContentSettingsData(TypedDict):
    filter_enabled: bool
    filter_locked: bool


class UserData(BaseObjectData):
    country: str
    display_name: str
    email: str
    explicit_content: ExplicitContentSettingsData
    external_urls: ExternalUrlsData
    followers: FollowersData
    images: list[ImageData]
    product: str


# PLAYLISTS API

class PlaylistSnapshotId(TypedDict):
    snapshot_id: str


class PlaylistTrackData(BaseObjectData):
    added_at: str
    added_by: UserData
    is_local: bool
    primary_color: Any
    video_thumbnail: Any
    track: TrackData


class PlaylistTrackRefData(TypedDict):
    href: str
    total: int


class SimplePlaylistData(BaseObjectData):
    collaborative: bool
    description: Optional[str]
    external_urls: ExternalUrlsData
    images: list[ImageData]
    owner: UserData
    primary_color: Optional[str]
    public: Optional[bool]
    snapshot_id: str
    tracks: PlaylistTrackRefData


class PlaylistData(BaseObjectData):
    collaborative: bool
    description: Optional[str]
    external_urls: ExternalUrlsData
    followers: FollowersData
    images: list[ImageData]
    owner: UserData
    primary_color: Optional[str]
    public: Optional[bool]
    snapshot_id: str
    tracks: PagingObjectData


class FeaturedPlaylistsData(TypedDict):
    message: str
    playlists: PagingObjectData


class CategoryPlaylistsData(TypedDict):
    playlists: PagingObjectData


# CATEGORY API

class CategoryData(TypedDict):
    href: str
    icons: list[ImageData]
    id: str
    name: str


class MultipleCategoriesData(TypedDict):
    categories: PagingObjectData


# GENRE API

class RecommendationGenresData(TypedDict):
    genres: list[str]


# PLAYER API

class DeviceData(TypedDict):
    id: str
    is_active: bool
    is_private_session: bool
    is_restricted: bool
    name: str
    type: str
    volume_percent: int


class MultipleDevicesData(TypedDict):
    devices: list[DeviceData]


class ActionsData(TypedDict):
    interrupting_playback: bool
    pausing: bool
    resuming: bool
    seeking: bool
    skipping_next: bool
    skipping_prev: bool
    toggling_repeat_context: bool
    toggling_repeat_track: bool
    toggling_shuffle: bool
    transferring_playback: bool


class ContextData(TypedDict):
    external_urls: ExternalUrlsData
    href: str
    type: str
    uri: str


class PlaybackStateData(TypedDict):
    actions: ActionsData
    context: ContextData
    currently_playing_type: str
    device: DeviceData
    is_playing: bool
    item: Optional[TrackData]
    progress_ms: int
    repeat_state: str
    shuffle_state: str
    timestamp: int


class CurrentlyPlayingData(TypedDict):
    context: ContextData
    currently_playing_type: str
    is_playing: bool
    item: Optional[TrackData]
    progress_ms: int
    timestamp: int


# MARKETS API

class AvailableMarketsData(TypedDict):
    markets: list[str]


# TOKENS

class ClientCredentialsData(TypedDict):
    access_token: str
    token_type: str
    expires_in: int


class UserCredentialsData(TypedDict):
    access_token: str
    token_type: str
    expires_in: int
    scope: str
    refresh_token: str
